import pkg_resources

from .dialog import dialogus
from .prompt import interrogatio

__all__ = ('dialogus', 'interrogatio')

try:
    __version__ = pkg_resources.get_distribution('interrogatio').version

    __version_info__ = tuple(
        [int(num) if num.isdigit() else
        num for num in __version__.replace('-', '.', 1).split('.')])

except:
    __version__ = '1.0.3'
    __version_info__ = (1, 0, 3)



from prompt_toolkit.application import Application
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings
from prompt_toolkit.key_binding.defaults import load_key_bindings
from prompt_toolkit.layout import HorizontalAlign, Layout
from prompt_toolkit.shortcuts import print_formatted_text

from .dialog import show_dialog
from .handlers import get_instance
from .themes import for_prompt, set_theme
from .utils import validate_questions
from .validators import Validator

__all__ = ['interrogatio']

def interrogatio(questions, theme='default'):
    """
    Prompts user for inputs as defined in the questions parameter and returns
    a dictionary with the answers.

    :param questions: a list of questions.
    :type questions: list
    :param theme: the name of the theme to use.
    :type theme: string

    :return: a dictionary with the answers.
    :rtype: dict

    :raise InvalidQuestionError: if there is an error in the question
                                 definition.
    :raise ThemeNotFoundError: if the specified theme does not exists.

    Usage:

    .. code-block:: python

        from interrogatio import interrogatio
        questions = [
            {
                'name': 'name',
                'type': 'input',
                'message': 'What is your name'
            },
            {
                'name': 'favorite_pet',
                'type': 'input',
                'message': 'What is your favorite pet'
            }
        ]
        answers = interrogatio(questions, theme='purple')
    """
    set_theme(theme)
    answers = {}
    validate_questions(questions)
    for q in questions:
        handler = get_instance(q)
        l = handler.get_layout()
        l.align = HorizontalAlign.LEFT

        bindings = [load_key_bindings()]

        handler_bindings = handler.get_keybindings()

        if handler_bindings:
            bindings.append(handler_bindings)

        app = Application(
            layout=Layout(l),
            key_bindings=merge_key_bindings(bindings),
            style=for_prompt())

        while True:
            app.run()
            if handler.is_valid():
                answers.update(handler.get_answer())
                break
            else:
                print_formatted_text(
                    FormattedText([
                        ('class:error', handler.errors[0])
                    ]),
                    style=for_prompt()
                )
    return answers



import abc

import six

from ..core.exceptions import AlreadyRegisteredError, ValidationError

__all__ = [
    'QHandler',
    'register',
    'get_instance',
    'get_registered'
]


class QHandlersRegistry(dict):
    def register(self, alias, clazz):
        if alias in self:
            raise AlreadyRegisteredError('alias {} already exists'.format(
                alias))
        self[alias] = clazz

    def get_registered(self):
        return list(self.keys())

    def get_instance(self, question):
        qtype = question['type']
        clazz = self[qtype]
        return clazz(question)

_registry = QHandlersRegistry()


def register(alias, clazz):
    """
    Register a question handler to use with interrogatio.
    Each question handler is identified by a unique alias.
    This alias is used in question definition (type) to invoke a specific
    handler.


    :param alias: a unique alias to indentify a question handler.
    :type alias: str

    :param clazz: a QHandler concrete implementation.
    :type clazz: class
    """
    _registry.register(alias, clazz)


def get_instance(question):
    """
    Returns an instance of a concrete subclass of QHandler initialized with
    a question. The subclass of QHandler is identified by the ``type`` field
    of the question.

    :param question: a question to handle.
    :type question: dict

    :return: an instance of the corresponding QHandler
    :rtype: QHandler subclass
    """
    return _registry.get_instance(question)


def get_registered():
    """
    Returns a list of registered QHandlers.

    :return: list of aliases of the registered QHandlers.
    :rtype: list
    """
    return _registry.get_registered()


class QHandler(six.with_metaclass(abc.ABCMeta, object)):
    """
    ABC for question handlers.
    Each question handler must subclass this class.
    """

    def __init__(self, question):
        self._question = question
        self._widget = None
        self._errors = []

    @property
    def errors(self):
        """
        This property holds a list of validation error messages.
        The list is filled after a call to the is_valid method.

        :return: list of error messages.
        :rtype: list
        """
        return self._errors

    @abc.abstractmethod
    def get_layout(self):
        """
        Returns the UI layout of the question. It must returns a
        python-prompt-toolkit layout container.

        Subclasses must implement this method.

        :return: the UI layout of the question.
        :rtype: :class:`~prompt_toolkit.layout.Layout`
        """


    @abc.abstractmethod
    def get_value(self):
        """
        Returns the ``value`` part of the answer.

        Subclasses must implement this method.

        :return: the ``value`` part of the answer.
        :rtype: str
        """

    @abc.abstractmethod
    def get_widget_init_kwargs(self):
        """
        Returns the keyword arguments needed to instantiate the widget.

        Subclasses must implement this method.

        :return: a dictionary containing the keyword arguments.
        :rtype: dict
        """

    def get_init_extra_args(self):
        """
        Returns extra arguments needed to instantiate a QHandler.
        The extra arguments are represented as a dictionary within the
        question object under the key ``extra_args``.

        :return: a dictionary containing the initialization extra arguments.
        :rtype: dict
        """
        return self._question.get('extra_args', dict())

    @abc.abstractmethod
    def get_widget_class(self):
        """
        Returns the widget class for this QHandler.

        :return: a widget class.
        :rtype: class
        """

    def get_widget(self):
        """
        Returns the widget instance for this QHandler.
        If the widget has not been already created, this method creates it
        before returns.
        """
        if not self._widget:
            clazz = self.get_widget_class()
            self._widget = clazz(**self.get_widget_init_kwargs())
        return self._widget

    @abc.abstractmethod
    def get_keybindings(self):
        """
        Returns a KeyBindings object to add custom keybindings to this
        QHandler.
        """

    def get_answer(self):
        """
        Returns dictionary with the question variable as key and the answer
        as the value.
        """
        return {self._question['name']: self.get_value()}

    def get_variable_name(self):
        """
        Returns the name of the variable of this question.
        """
        return self._question['name']

    def is_valid(self):
        """
        Apply any speficied validator to the answer and return True if the
        input is valid otherwise False.
        If the answer isn't valid, it also set the errors property to a list
        of error messages.
        """
        validators = self._question.get('validators', [])
        self._errors = []
        for validator in validators:
            try:
                validator.validate(self.get_value())
            except ValidationError as ve:
                self._errors.append(ve.message)
        return not self._errors



import abc
import six


from ..core.exceptions import AlreadyRegisteredError


__all__ = [
    'Validator',
    'register',
    'get_registered',
    'get_instance'
]


class ValidatorsRegistry(dict):
    def register(self, alias, clazz):
        if alias in self:
            raise AlreadyRegisteredError(
                'validator {} already registered'.format(alias))
        self[alias] = clazz

    def get_registered(self):
        return list(self.keys())

    def get_instance(self, v):
        clazz = self[v['name']]
        if 'args' in v:
            return clazz(**v['args'])
        return clazz()

_registry = ValidatorsRegistry()


def register(alias, clazz):
    """
    Register a validator to use with interrogatio.
    Each validator is identified by a unique alias.
    This alias can be used in question definition to validater answers.


    :param alias: a unique alias to indentify a validator.
    :type alias: str

    :param clazz: a Validator concrete implementation.
    :type clazz: class
    """
    _registry.register(alias, clazz)

def get_registered():
    """
    Returns a list of registered Validators.

    :return: list of aliases of the registered Validators.
    :rtype: list
    """
    return _registry.get_registered()

def get_instance(v):
    """
    Returns an instance of a concrete subclass of Validator.

    :param v: a validator definition object.
    :type v: dict

    :return: an instance of the corresponding Validator.
    :rtype: Validator subclass
    """
    return _registry.get_instance(v)


class Validator(six.with_metaclass(abc.ABCMeta, object)):
    """
    Abstract class for validators.
    """
    def __init__(self, message='invalid input'):
        self.message = message

    @abc.abstractmethod
    def validate(self, value):
        """
        Abstract method.
        Subclasses must implement this method with the validation logic.

        :param value: the value to validate.
        :type value: str

        :raises:
            ValidationError: if the provided input is invalid.
        """



class AlreadyRegisteredError(Exception):
    pass


class InvalidQuestionError(Exception):
    pass


class ThemeNotFoundError(Exception):
    pass


class ValidationError(Exception):
    """
    Exception raised when validation fails.
    """
    def __init__(self, message):
        super(ValidationError, self).__init__()
        self._message = message

    def __str__(self):
        return self._message

    @property
    def message(self):
        return self._message



import json

from prompt_toolkit.styles import Style, default_ui_style, merge_styles

from ..core.exceptions import AlreadyRegisteredError, ThemeNotFoundError

__all__ = [
    'Theme',
    'register',
    'for_dialog',
    'for_prompt',
    'set_theme'
]


class Theme:

    def __init__(self):
        self._prompt_styles = dict()
        self._dialog_styles = dict()
        self._name = ''

    def load(self, filename):
        with open(filename, 'r') as f:
            tmp = json.load(f)
            self._name = tmp['name']
            self._prompt_styles = tmp['prompt']
            self._dialog_styles = tmp['dialog']

    def for_prompt(self):
        return merge_styles([
            default_ui_style(),
            Style(list(self._prompt_styles.items()))
        ])
    def for_dialog(self):
        return merge_styles([
            default_ui_style(),
            Style(list(self._dialog_styles.items()))
        ])

    def save(self, filename):
        with open(filename, 'w') as f:
            json.dump({
                'name': self._name,
                'prompt': self._prompt_styles,
                'dialog': self._dialog_styles
            }, f, indent=2)



class ThemeRegistry(object):

    def __init__(self):
        self._themes = {}
        self._current_theme = 'default'

    def register(self, alias, theme):
        assert isinstance(theme, Theme)
        if alias in self._themes:
            raise AlreadyRegisteredError(
                'theme {} already registered'.format(alias))
        self._themes[alias] = theme


    def set_current(self, alias):
        if alias not in self._themes:
            raise ThemeNotFoundError('theme {} not registered'.format(alias))
        self._current_theme = alias

    def get_current(self):
        return self._themes[self._current_theme]


_registry = ThemeRegistry()


def register(alias, theme):
    _registry.register(alias, theme)

def for_dialog():
    return _registry.get_current().for_dialog()

def for_prompt():
    return _registry.get_current().for_prompt()

def set_theme(alias):
    _registry.set_current(alias)



# pylint: disable=unused-argument
import string

from prompt_toolkit.application.current import get_app
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import HorizontalAlign, HSplit, VSplit
from prompt_toolkit.layout.dimension import Dimension as D
from prompt_toolkit.widgets import Label, TextArea, Button, Box

from ..widgets import SelectMany, SelectOne, MaskedInput, DateRange
from .base import QHandler, register


class StringHandler(QHandler):

    def get_widget_class(self):
        return TextArea

    def get_widget_init_kwargs(self):
        kwargs = dict(
            multiline=self._question.get('multiline', False),
            style='class:input.answer'
        )
        if 'default' in self._question:
            kwargs['text'] = self._question['default']
        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        widget.buffer.cursor_position = len(widget.text)
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0, 
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:input.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0, 
                Label(
                    self._question['description'],
                    style='class:input.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)


    def get_keybindings(self):
        bindings = KeyBindings()

        @bindings.add(Keys.ControlC)
        def _ctrl_c(event):
            get_app().exit(result=False)

        @bindings.add(Keys.Enter)
        def _enter(event):
            get_app().exit(result=True)

        return bindings

    def get_value(self):
        return self.get_widget().text


register('input', StringHandler)



class PasswordHandler(QHandler):

    def get_widget_class(self):
        return TextArea

    def get_widget_init_kwargs(self):
        kwargs = dict(
            multiline=False,
            style='class:password.answer',
            password=True
        )
        if 'default' in self._question:
            kwargs['text'] = self._question['default']
        return kwargs

    def get_value(self):
        return self.get_widget().text

    def get_layout(self):
        msg = '{}{}'.format(
            self._question['message'],
            self._question.get('question_mark', ' ?')
        )
        widget = self.get_widget()
        widget.buffer.cursor_position = len(widget.text)

        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0, 
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:password.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0, 
                Label(
                    self._question['description'],
                    style='class:password.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)


    def get_keybindings(self):
        bindings = KeyBindings()

        @bindings.add(Keys.ControlC)
        def _ctrl_c(event):
            get_app().exit(exception=KeyboardInterrupt)

        @bindings.add(Keys.Enter)
        def _enter(event):
            get_app().exit(result=self.get_answer())

        return bindings

register('password', PasswordHandler)


class SelectOneHandler(QHandler):

    def get_widget_class(self):
        return SelectOne

    def get_value(self):
        return self.get_widget().current_value

    def get_widget_init_kwargs(self):
        kwargs = dict(
            values=self._question['values'],
            style='class:selectone.answer'
        )
        if 'default' in self._question:
            kwargs['default'] = self._question['default']

        return kwargs

    def get_layout(self):
        vsplit_components = [self.get_widget()]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0, 
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:selectone.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0, 
                Label(
                    self._question['description'],
                    style='class:selectone.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)


    def get_keybindings(self):

        bindings = KeyBindings()

        @bindings.add(Keys.ControlC)
        def _ctrl_c(event):
            get_app().exit(exception=KeyboardInterrupt)

        def accept_handler(value):
            get_app().exit(result=self.get_answer())

        self.get_widget().accept_handler = accept_handler

        return bindings

register('selectone', SelectOneHandler)


class SelectManyHandler(QHandler):


    def get_widget_class(self):
        return SelectMany


    def get_value(self):
        return list(self.get_widget().checked)

    def get_widget_init_kwargs(self):
        kwargs = dict(
            values=self._question['values'],
            style='class:selectmany.answer'
        )
        if 'checked' in self._question:
            kwargs['checked'] = set(self._question['checked'])
        if 'default' in self._question:
            kwargs['default'] = self._question['default']

        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0, 
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:selectone.question',
                ),
            )

        btn_all = Button('All', widget.select_all)
        btn_none = Button('None', widget.select_none)
        buttons = Box(
            body=VSplit([btn_all, btn_none], padding=1),
            height=D(min=1, max=3, preferred=3),
            padding_left=0,
        )

        hsplit_components = [buttons, VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0, 
                Label(
                    self._question['description'],
                    style='class:selectone.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_keybindings(self):

        bindings = KeyBindings()

        @bindings.add(Keys.ControlC)
        def _ctrl_c(event):
            get_app().exit(exception=KeyboardInterrupt)

        def accept_handler(value):
            get_app().exit(result=self.get_answer())

        self.get_widget().accept_handler = accept_handler

        return bindings

register('selectmany', SelectManyHandler)

# class TextHandler(QHandler):

#     def get_widget_class(self):
#         return TextArea


#     def get_widget_init_kwargs(self):
#         kwargs = dict(
#             multiline=True,
#             height=4,
#             style='class:text.answer'
#         )
#         if 'default' in self._question:
#             kwargs['text'] = self._question['default']
#         return kwargs

#     def get_value(self):
#         return self.get_widget().text

#     def get_layout(self):
#         msg = '{}{}'.format(
#             self._question['message'],
#             self._question.get('question_mark', ' ?')
#         )
#         extra_args = self.get_init_extra_args()
#         if 'rows' in extra_args:
#             self.get_widget().height = int(extra_args['rows'])
#         return VSplit(
#             [
#                 Label(msg,
#                       dont_extend_width=True,
#                       style='class:text.question'),
#                 self.get_widget()
#             ],
#             padding=1)

#     def get_keybindings(self):
#         bindings = KeyBindings()

#         @bindings.add(Keys.ControlC)
#         def _ctrl_c(event):
#             get_app().exit(exception=KeyboardInterrupt)

#         @bindings.add(Keys.ControlX)
#         def _enter(event):
#             get_app().exit(result=self.get_answer())

#         return bindings


# class PathHandler(StringHandler):
#     def get_widget(self):
#         widget = super(PathHandler, self).get_widget()
#         widget.completer = PathCompleter()
#         return widget

# register('path', PathHandler)

# class RePasswordHandler(QHandler):

#     ALIAS = 'repassword'


#     def __init__(self, *args, **kwargs):
#         super(RePasswordHandler, self).__init__(*args, **kwargs)
#         self._rewidget = TextArea(**self.get_widget_init_kwargs())
#         self._rewidget.buffer.cursor_position = len(self._rewidget.text)


#     def get_widget_class(self):
#         return TextArea

#     def get_widget_init_kwargs(self):
#         kwargs = dict(
#             multiline=False,
#             style='class:repassword.answer',
#             password=True
#         )
#         if 'default' in self._question:
#             kwargs['text'] = self._question['default']
#         return kwargs

#     def get_value(self):
#         return self.get_widget().text

#     def get_layout(self):
#         msg = '{}{}'.format(
#             self._question['message'],
#             self._question.get('question_mark', ' ?')
#         )
#         extra_args = self.get_init_extra_args()
#         remsg = '{}{}'.format(
#             extra_args.get('remessage', 'Please repeat to confirm'),
#             self._question.get('question_mark', ' ?')
#         )
#         lbl_msg = Label(msg,
#                         dont_extend_width=True,
#                         style='class:repassword.question')
#         lbl_remsg = Label(remsg,
#                           dont_extend_width=True,
#                           style='class:repassword.question')
#         return HSplit(
#             [
#                 VSplit([lbl_msg, self.get_widget()], padding=1),
#                 VSplit([lbl_remsg, self._rewidget], padding=1)
#             ]
#         )

#     def get_keybindings(self):
#         bindings = KeyBindings()

#         @bindings.add(Keys.ControlC)
#         def _ctrl_c(event):
#             get_app().exit(exception=KeyboardInterrupt)


#         @bindings.add('tab')
#         def _tab(event):
#             get_app().layout.focus_next()

#         @bindings.add('s-tab')
#         def _stab(event):
#             get_app().layout.focus_previous()

#         @bindings.add(Keys.Enter)
#         def _enter(event):
#             if get_app().layout.has_focus(self._rewidget):
#                 get_app().exit(result=self.get_answer())
#             else:
#                 get_app().layout.focus_next()

#         return bindings

#     def is_valid(self):
#         extra_args = self.get_init_extra_args()
#         super(RePasswordHandler, self).is_valid()
#         if self.get_widget().text != self._rewidget.text:
#             self._errors.append(
#                 extra_args.get(
#                     'reerror',
#                     'Password and repeat password doesn\'t match'
#                 )
#             )
#         return not self.errors

# register('repassword', RePasswordHandler)


class MaskedInputHandler(QHandler):

    def get_widget_class(self):
        return MaskedInput

    def get_widget_init_kwargs(self):
        kwargs = dict(
            style='class:input.answer',
            mask=self._question['mask'],
        )
        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0, 
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:input.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0, 
                Label(
                    self._question['description'],
                    style='class:input.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_keybindings(self):
        bindings = KeyBindings()

        @bindings.add(Keys.ControlC)
        def _ctrl_c(event):
            get_app().exit(result=False)

        @bindings.add(Keys.Enter)
        def _enter(event):
            get_app().exit(result=True)

        return bindings

    def get_value(self):
        return self.get_widget().value


register('maskedinput', MaskedInputHandler)



class DateHandler(QHandler):

    def get_widget_class(self):
        return MaskedInput

    def get_widget_init_kwargs(self):
        kwargs = dict(
            style='class:input.answer',
            mask='____-__-__',
            allowed_chars=string.digits,
        )
        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0, 
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:input.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0, 
                Label(
                    self._question['description'],
                    style='class:input.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_keybindings(self):
        bindings = KeyBindings()

        @bindings.add(Keys.ControlC)
        def _ctrl_c(event):
            get_app().exit(result=False)

        @bindings.add(Keys.Enter)
        def _enter(event):
            get_app().exit(result=True)

        return bindings

    def get_value(self):
        return self.get_widget().value


register('date', DateHandler)

class DateRangeHandler(QHandler):

    def get_widget_class(self):
        return DateRange

    def get_widget_init_kwargs(self):
        kwargs = dict(
            style='class:input.answer',
        )
        if 'from_label' in self._question:
            kwargs['from_label'] = self._question['from_label']
        if 'to_label' in self._question:
            kwargs['to_label'] = self._question['to_label']     
        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0, 
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:input.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0, 
                Label(
                    self._question['description'],
                    style='class:input.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)


    def get_keybindings(self):
        bindings = KeyBindings()

        @bindings.add(Keys.ControlC)
        def _ctrl_c(event):
            get_app().exit(result=False)

        @bindings.add(Keys.Enter)
        def _enter(event):
            get_app().exit(result=True)

        return bindings

    def get_value(self):
        return self.get_widget().value


register('daterange', DateRangeHandler)


def to_style_token(fg='', bg='', attr=''):
    tokens = []
    if bg:
        tokens.append('bg:{}'.format(bg))
    if fg:
        tokens.append(fg)
    if attr:
        tokens.append(attr)
    return ' '.join(tokens)



from prompt_toolkit.shortcuts import button_dialog as pt_button_dialog
from prompt_toolkit.shortcuts import input_dialog as pt_input_dialog
from prompt_toolkit.shortcuts import message_dialog as pt_message_dialog
from prompt_toolkit.shortcuts import progress_dialog as pt_progress_dialog
from prompt_toolkit.shortcuts import radiolist_dialog as pt_radiolist_dialog
from prompt_toolkit.shortcuts import yes_no_dialog as pt_yes_no_dialog

from ..themes import for_dialog

__all__ = [
    'yes_no_dialog',
    'button_dialog',
    'input_dialog',
    'message_dialog',
    'radiolist_dialog',
    'progress_dialog',
]

def yes_no_dialog(title='', text='', yes_text='Yes', no_text='No',
                  style=None, async_=False):
    style = style or for_dialog()
    return pt_yes_no_dialog(title, text, yes_text, no_text,
                            style, async_)

def button_dialog(title='', # pylint: disable=dangerous-default-value
                  text='',
                  buttons=[],
                  style=None,
                  async_=False):
    style = style or for_dialog()
    return pt_button_dialog(title, text, buttons, style, async_)

def input_dialog(title='', text='', ok_text='OK', cancel_text='Cancel',
                 completer=None, password=False, style=None,
                 async_=False):
    style = style or for_dialog()
    return pt_input_dialog(title, text, ok_text, cancel_text,
                           completer, password, style, async_)

def message_dialog(title='', text='', ok_text='Ok',
                   style=None, async_=False):
    style = style or for_dialog()
    return pt_message_dialog(title, text, ok_text, style, async_)

def radiolist_dialog(title='', text='', ok_text='Ok', cancel_text='Cancel',
                     values=None, style=None,
                     async_=False):
    style = style or for_dialog()
    return pt_radiolist_dialog(title, text, ok_text, cancel_text, values,
                               style, async_)

def progress_dialog(title='', text='', run_callback=None,
                    style=None, async_=False):
    style = style or for_dialog()
    return pt_progress_dialog(title, text, run_callback, style, async_)



import argparse
import json
import sys

from functools import partial

try:
    import yaml
    FORMAT_CHOICES = ['json', 'yaml']
except ImportError:
    yaml = None
    FORMAT_CHOICES = ['json']

from . import interrogatio, dialogus


def _load_questions(args):
    with args.input as f:
        return args.deserialize(f)


def _write_answers(args, answers):
    with args.output as f:
        args.serialize(answers, f)


def main_dialogus():
    parser = argparse.ArgumentParser(description='dialogus')
    parser.add_argument('--input',
                        '-i',
                        type=argparse.FileType('r'),
                        required=True)
    parser.add_argument('--output',
                        '-o',
                        type=argparse.FileType('w'),
                        default=sys.stdout)

    parser.add_argument('--theme',
                        '-t',
                        default='default')

    parser.add_argument('--title')
    parser.add_argument('--confirm')
    parser.add_argument('--cancel')

    parser.add_argument('--input-format',
                        choices=FORMAT_CHOICES,
                        default='json')
    parser.add_argument('--output-format',
                        choices=FORMAT_CHOICES,
                        default='json')

    args = parser.parse_args()

    if args.input_format == 'yaml':
        args.deserialize = partial(yaml.load, Loader=yaml.FullLoader)
        args.serialize = yaml.dump
    else:
        args.deserialize = json.load
        args.serialize = json.dump

    kwargs = dict(
        theme=args.theme
    )
    if args.title:
        kwargs['title'] = args.title
    if args.confirm:
        kwargs['confirm'] = args.confirm
    if args.cancel:
        kwargs['cancel'] = args.cancel
    _write_answers(args, dialogus(_load_questions(args), **kwargs))


def main_interrogatio():
    parser = argparse.ArgumentParser(description='interrogatio')
    parser.add_argument('--input',
                        '-i',
                        type=argparse.FileType('r'),
                        required=True)
    parser.add_argument('--output',
                        '-o',
                        type=argparse.FileType('w'),
                        default=sys.stdout)

    parser.add_argument('--theme',
                        '-t',
                        default='default')

    parser.add_argument('--input-format',
                        choices=FORMAT_CHOICES,
                        default='json')
    parser.add_argument('--output-format',
                        choices=FORMAT_CHOICES,
                        default='json')

    args = parser.parse_args()

    if args.input_format == 'yaml':
        args.deserialize = partial(yaml.load, Loader=yaml.FullLoader)
        args.serialize = yaml.dump
    else:
        args.deserialize = json.load
        args.serialize = json.dump

    _write_answers(args, interrogatio(_load_questions(args),
                                      theme=args.theme))



# pylint: disable=unused-argument
import string

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout.containers import Window, VSplit, HSplit
from prompt_toolkit.layout.controls import FormattedTextControl, BufferControl
from prompt_toolkit.layout.dimension import Dimension as D
from prompt_toolkit.layout.margins import ScrollbarMargin
from prompt_toolkit.layout.processors import Processor
from prompt_toolkit.widgets import Label, TextArea
from prompt_toolkit.mouse_events import MouseEventType


class SelectOne(object):
    def __init__(self, values=None, default=None,
                 accept_handler=None, style=''):
        assert isinstance(values, list)
        assert all(isinstance(i, tuple) and len(i) == 2
                   for i in values)

        self.values = values or []
        self.current_value = values[0][0]
        self._selected_index = 0
        self.accept_handler = accept_handler

        for i in range(len(self.values)):
            if self.values[i][0] == default:
                self._selected_index = i
                self.current_value = self.values[i][0]
                break


        # Key bindings.
        kb = KeyBindings()

        @kb.add('up')
        def _(event):
            self._selected_index = max(0, self._selected_index - 1)

        @kb.add('down')
        def _(event):
            self._selected_index = min(
                len(self.values) - 1, self._selected_index + 1)

        @kb.add('pageup')
        def _(event):
            w = event.app.layout.current_window
            self._selected_index = max(
                0,
                self._selected_index - len(w.render_info.displayed_lines)
            )

        @kb.add('pagedown')
        def _(event):
            w = event.app.layout.current_window
            self._selected_index = min(
                len(self.values) - 1,
                self._selected_index + len(w.render_info.displayed_lines)
            )

        @kb.add(' ')
        def _(event):
            self.current_value = self.values[self._selected_index][0]

        @kb.add('enter')
        def _(event):
            if self.accept_handler:
                self.accept_handler(self.current_value)

        @kb.add(Keys.Any)
        def _(event):
            # We first check values after the selected value, then all values.
            for value in self.values[self._selected_index + 1:] + self.values:
                if value[1].startswith(event.data):
                    self._selected_index = self.values.index(value)
                    return

        # Control and window.
        self.control = FormattedTextControl(
            lambda: self._get_text_fragments(style),
            key_bindings=kb,
            focusable=True)

        self.window = Window(
            content=self.control,
            style='class:radio-list',
            right_margins=[ScrollbarMargin(display_arrows=True)],
            dont_extend_height=True)

    def _get_text_fragments(self, out_style):
        def mouse_handler(mouse_event):
            """
            Set `_selected_index` and `current_value` according to the y
            position of the mouse click event.
            """
            if mouse_event.event_type == MouseEventType.MOUSE_UP:
                self._selected_index = mouse_event.position.y
                self.current_value = self.values[self._selected_index][0]

        result = []
        for i, value in enumerate(self.values):
            checked = (value[0] == self.current_value)
            selected = (i == self._selected_index)
            style = out_style
            if checked:
                style += ' class:radio-checked'
            if selected:
                style += ' class:radio-selected'

            result.append((style, '('))

            if selected:
                result.append(('[SetCursorPosition]', ''))

            if checked:
                result.append((style, '*'))
            else:
                result.append((style, ' '))

            result.append((style, ')'))
            result.append((out_style + ' class:radio', ' '))
            result.extend(
                to_formatted_text(value[1], style=out_style + ' class:radio'))
            result.append(('', '\n'))

        # Add mouse handler to all fragments.
        for i, fragment in enumerate(result):
            result[i] = (fragment[0], fragment[1], mouse_handler)

        result.pop()  # Remove last newline.
        return result

    def __pt_container__(self):
        return self.window




class SelectMany(object):
    def __init__(self, values=None, checked=None, default=None,
                 accept_handler=None, style=''):
        assert isinstance(values, list)
        assert all(isinstance(i, tuple) and len(i) == 2
                   for i in values)

        self.values = values
        self.checked = checked or set()
        self._selected_index = 0
        self.accept_handler = accept_handler

        # for i in range(len(self.values)):
        #     if self.values[i][0] == default:
        #         self._selected_index = i
        #         self.current_value = self.values[i][0]
        #         break


        # Key bindings.
        kb = KeyBindings()

        @kb.add('up')
        def _(event):
            self._selected_index = max(0, self._selected_index - 1)

        @kb.add('down')
        def _(event):
            self._selected_index = min(
                len(self.values) - 1, self._selected_index + 1)

        @kb.add('pageup')
        def _(event):
            w = event.app.layout.current_window
            self._selected_index = max(
                0,
                self._selected_index - len(w.render_info.displayed_lines)
            )

        @kb.add('pagedown')
        def _(event):
            w = event.app.layout.current_window
            self._selected_index = min(
                len(self.values) - 1,
                self._selected_index + len(w.render_info.displayed_lines)
            )

        @kb.add(' ')
        def _(event):
            if self.values[self._selected_index][0] in self.checked:
                self.checked.remove(self.values[self._selected_index][0])
            else:
                self.checked.add(self.values[self._selected_index][0])


        @kb.add('enter')
        def _(event):
            if self.accept_handler:
                self.accept_handler(list(self.checked))

        @kb.add(Keys.Any)
        def _(event):
            # We first check values after the selected value, then all values.
            for value in self.values[self._selected_index + 1:] + self.values:
                if value[1].startswith(event.data):
                    self._selected_index = self.values.index(value)
                    return

        # Control and window.
        self.control = FormattedTextControl(
            lambda: self._get_text_fragments(style),
            key_bindings=kb,
            focusable=True)

        self.window = Window(
            content=self.control,
            style='class:checkbox-list',
            right_margins=[ScrollbarMargin(display_arrows=True)],
            dont_extend_height=True)

    def select_all(self):
        for v in self.values:
            self.checked.add(v[0])

    def select_none(self):
        self.checked.clear()

    def _get_text_fragments(self, out_style):
        def mouse_handler(mouse_event):
            """
            Set `_selected_index` and `current_value` according to the y
            position of the mouse click event.
            """
            if mouse_event.event_type == MouseEventType.MOUSE_UP:
                self._selected_index = mouse_event.position.y
                if self.values[self._selected_index][0] in self.checked:
                    self.checked.remove(self.values[self._selected_index][0])
                else:
                    self.checked.add(self.values[self._selected_index][0])

        result = []
        for i, value in enumerate(self.values):
            checked = (value[0] in self.checked)
            selected = (i == self._selected_index)
            style = out_style
            if checked:
                style += ' class:checkbox-checked'
            if selected:
                style += ' class:checkbox-selected'

            result.append((style, '['))

            if selected:
                result.append(('[SetCursorPosition]', ''))

            if checked:
                result.append((style, '*'))
            else:
                result.append((style, ' '))

            result.append((style, ']'))
            result.append((out_style + ' class:checkbox', ' '))
            result.extend(
                to_formatted_text(value[1],
                                  style=out_style + ' class:checkbox'))
            result.append(('', '\n'))

        # Add mouse handler to all fragments.
        for i, fragment in enumerate(result):
            result[i] = (fragment[0], fragment[1], mouse_handler)

        result.pop()  # Remove last newline.
        return result

    def __pt_container__(self):
        return self.window



class FixedLengthBuffer(Buffer):
    def __init__(self, **kwargs):
        self._max_length = kwargs.pop('max_length')
        self._focus_next = kwargs.pop('focus_next')
        self._allowed_chars = kwargs.pop('allowed_chars')
        super().__init__(**kwargs)

    def _is_input_allowed(self, data):
        if not self._allowed_chars:
            return True
        return all([char in self._allowed_chars for char in data])

    def insert_text(
        self,
        data,
        overwrite=False,
        move_cursor=True,
        fire_event=True,
    ):  
        if len(self.document.text) + len(data) <= self._max_length:
            if self._is_input_allowed(data):
                super().insert_text(data, overwrite, move_cursor, fire_event)
        if self.cursor_position == self._max_length:
            self.cursor_position -= 1
            if self._focus_next:
                get_app().layout.focus_next()



class FixedLengthTextArea(TextArea):
 
    def __init__(
        self,
        text='',
        width=None,
        height=None,
        max_length=None,
        focus_next=True,
        allowed_chars=None,
        style=None,
    ):


        self.buffer = FixedLengthBuffer(
            document=Document(text, 0),
            multiline=False,
            max_length=max_length,
            focus_next=focus_next,
            allowed_chars=allowed_chars,
            accept_handler=None,
        )

        self.control = BufferControl(
            buffer=self.buffer,
            focusable=True,
            focus_on_click=True,
        )

        height = D.exact(1)
        left_margins = []
        right_margins = []

        # style = "class:text-area "

        self.window = Window(
            height=height,
            width=width,
            content=self.control,
            style=style,
            wrap_lines=False,
        )


class MaskedInput(VSplit):

    def __init__(
        self,
        mask,
        placeholder='_',
        style=None,
        value=None,
        allowed_chars=None,
        accept_handler=None,
    ):
        self._mask = mask
        self._placeholder = placeholder
        
        self._value = value
        self.accept_handler = accept_handler

        self.current_position = 0
        self._components = []

        # Key bindings.
        kb = KeyBindings()

        @kb.add('enter')
        def _(event):
            if self.accept_handler:
                self.accept_handler(self.value)

        size = 0
        for i in range(len(self._mask)):
            char = self._mask[i]
            if char == placeholder:
                size += 1
                continue
            not_latest_field = char in self._mask[i:]
            if size > 0:
                widget = FixedLengthTextArea(
                    width=size,
                    max_length=size,
                    style=style,
                    allowed_chars=allowed_chars,
                    focus_next=not_latest_field,
                )
                self._components.append(widget)
                size = 0
            self._components.append(Label(char, dont_extend_width=True))

        if size > 0:
            widget = FixedLengthTextArea(
                width=size,
                max_length=size,
                style=style,
                focus_next=False,
                allowed_chars=allowed_chars,
            )
            self._components.append(widget)       
        self._components.append(Label(''))

        super().__init__(self._components, key_bindings=kb)

    def _has_value(self):
        values = ''.join([
            cmp.text for cmp in self._components
            if isinstance(cmp, FixedLengthTextArea)
        ]).strip()
        return bool(values)

    @property
    def value(self):
        if self._has_value():
            return ''.join([cmp.text for cmp in self._components])
        return None



class DateRange(HSplit):

    def __init__(self, from_label='From: ', to_label='  to: ', accept_handler=None, style=None):

        self._from = MaskedInput(mask='____-__-__', allowed_chars=string.digits, style=style)
        self._to = MaskedInput(mask='____-__-__', allowed_chars=string.digits, style=style)

        self.accept_handler = accept_handler

        # Key bindings.
        kb = KeyBindings()

        @kb.add('enter')
        def _(event):
            if self.accept_handler:
                self.accept_handler(self.value)

        components = []

        if from_label:
            components.append(VSplit([
                Label(from_label, dont_extend_width=True),
                self._from,
            ]))
        else:
            components.append(self._from)

        if to_label:
            components.append(
                VSplit([
                    Label(to_label, dont_extend_width=True),
                    self._to,
                ])
            )
        else:
            components.append(self._to)

        super().__init__(components, key_bindings=kb)
    
    @property
    def value(self):
        return {
            'from': self._from.value,
            'to': self._to.value,
        }


import os
from .base import Theme, register


class DefaultTheme(Theme):
    def __init__(self):
        super(DefaultTheme, self).__init__()
        self.load(os.path.join(
            os.path.dirname(__file__),
            'theme_files',
            'default.json'
        ))

register('default', DefaultTheme())


class PurpleTheme(Theme):
    def __init__(self):
        super(PurpleTheme, self).__init__()
        self.load(os.path.join(
            os.path.dirname(__file__),
            'theme_files',
            'purple.json'
        ))

register('purple', PurpleTheme())



from .core.exceptions import InvalidQuestionError
from . import validators
from . import handlers


def _validate_validator_object(obj):
    if 'name' not in obj:
        raise InvalidQuestionError(
            'You must specify a name for the validator')

    if obj['name'] not in validators.get_registered():
        raise InvalidQuestionError(
            'Validator {} does not exists'.format(obj['name']))

    if 'args' in obj and not isinstance(obj['args'], dict):
        raise InvalidQuestionError('Validator arguments must be a dictionary')


def _validate_question(q):

    if 'name' not in q:
        raise InvalidQuestionError('You must specify a name for the question')

    if 'type' not in q:
        raise InvalidQuestionError('You must specify a question type')

    q_type = q['type']
    if q_type not in handlers.get_registered():
        raise InvalidQuestionError('Unsupported question type: {}'.format(
            q_type))

    if q_type == 'selectone':
        if 'values' not in q:
            raise InvalidQuestionError('You must specify at least one '
                                       'choice for type choice')
        values = q['values']
        if not values:
            raise InvalidQuestionError('You must specify at least one '
                                       'choice for type choice')
        if not isinstance(values, (list, tuple)):
            raise InvalidQuestionError('Choices must be a list or tuple of'
                                       ' tuples.')
        first_value = values[0]
        if not isinstance(first_value, (list, tuple)):
            raise InvalidQuestionError('Choices must be a list or tuple of'
                                       ' tuples.')
        if len(first_value) != 2:
            raise InvalidQuestionError('Every choice must be a tuple'
                                       ' (value, label)')

    if 'validators' in q:
        if not isinstance(q['validators'], (list, tuple)):
            raise InvalidQuestionError('Validators must be a list or tuple')

        validator_instances = []
        for v in q['validators']:
            if not isinstance(v, (validators.Validator, dict)):
                raise InvalidQuestionError(
                    'Validators must be a list of  '
                    'interrogatio.validators.Validator'
                    ' instances or a list of validator objects')

            if isinstance(v, dict):
                _validate_validator_object(v)
                v = validators.get_instance(v)
                validator_instances.append(v)
            else:
                validator_instances.append(v)
        q['validators'] = validator_instances


def validate_questions(questions):
    for q in questions:
        _validate_question(q)



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    long_description = f.read()


setup(
    name='interrogatio',
    version='1.0.3',
    description='A python library to prompt user '
                'for inputs in a terminal application.',
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        'prompt-toolkit>=2.0.8',
        'pygments>=2.3.1',
        'six>=1.12.0'
    ],
    extras_require={
        'yaml':  ['pyyaml>=5',],
        'zxcvbn': ['zxcvbn>=4.4.27'],
        'test':  ['isort', 'flake8', 'ipdb', 'pytest', 'pytest-cov', 'pytest-runner'],
    },
    entry_points={
        'console_scripts': [
            'interrogatio = interrogatio.main:main_interrogatio',
            'dialogus = interrogatio.main:main_dialogus'
        ]
    },
    author='Francesco Faraone',
    author_email='ffaraone@gmail.com',
    url='https://github.com/ffaraone/interrogatio',
    license='BSD 3-Clause',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Terminals'
    ],
    keywords='console input dialog curses prompt',
    include_package_data=True
)



import collections
import ipaddress
import re

from datetime import datetime

from urllib.parse import urlsplit, urlunsplit

from .base import Validator, register
from ..core.exceptions import ValidationError


__all__ = [
    'RequiredValidator',
    'RegexValidator',
    'URLValidator',
    'EmailValidator',
    'MinLengthValidator',
    'MaxLengthValidator',
    'NumberValidator',
    'IntegerValidator',
    'IPv4Validator',
    'RangeValidator',
    'MinValidator',
    'MaxValidator',
    'DateTimeValidator',
    'DateTimeRangeValidator',
]


def _validate_ipv4_address(value):
    try:
        ipaddress.IPv4Address(value)
    except ValueError:
        raise ValidationError('invalid IPv4 address')

def _validate_ipv6_address(value):
    try:
        ipaddress.IPv6Address(value)
    except ValueError:
        raise ValidationError('invalid IPv6 address')


def validate_ipv46_address(value):
    try:
        _validate_ipv4_address(value)
    except ValidationError:
        try:
            _validate_ipv6_address(value)
        except ValidationError:
            raise ValidationError(message='invalid IPv4 or IPv6 address')


class RequiredValidator(Validator):
    def __init__(self, message=None):
        """
        Initialise the ``required`` validator.

        :param message: the error message in case that validation fails.
        :type message: str
        """
        super(RequiredValidator, self).__init__(
            message=message or 'this field is required'
        )

    def validate(self, value):
        if isinstance(value, str):
            if value.strip():
                return
        elif isinstance(value, (list, tuple)):
            if len(value) > 0:
                return
        elif isinstance(value, dict):
            if all(value.values()):
                return
        raise ValidationError(self.message)


register('required', RequiredValidator)

class RegexValidator(Validator):
    def __init__(self, expr, message=None, inverse_match=None):
        """
        Initialise the ``regex`` validator.

        :param expr: the regex to match.
        :type expr: str
        :param message: the error message in case that validation fails.
        :type message: str
        :param inverse_match: invert the match of the expression.
        :type inverse_match: bool
        """
        super(RegexValidator, self).__init__(
            message=message or 'this field does not match {}'.format(expr)
        )
        self.regex = re.compile(expr)
        self.inverse_match = inverse_match

    def validate(self, value):
        regex_matches = self.regex.search(str(value))
        invalid_input = regex_matches \
                if self.inverse_match else not regex_matches
        if invalid_input:
            raise ValidationError(self.message)


register('regex', RegexValidator)

class EmailValidator(Validator):
    user_regex = re.compile(
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*\Z"
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|'
        r'\\[\001-\011\013\014\016-\177])*"\Z)',
        re.IGNORECASE)
    domain_regex = re.compile(
        r'((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+)'
        r'(?:[A-Z0-9-]{2,63}(?<!-))\Z',
        re.IGNORECASE)
    literal_regex = re.compile(
        r'\[([A-f0-9:\.]+)\]\Z',
        re.IGNORECASE)
    domain_whitelist = ['localhost']

    def __init__(self, whitelist=None, message=None):
        """
        Initialise the ``email`` validator.

        :param whitelist: an optiona whitelist of domain names to be
                          considered valid.
        :type whitelist: list
        :param message: the error message in case that validation fails.
        :type message: str
        """
        super(EmailValidator, self).__init__(
            message=message or 'this field must be an email address'
        )
        if whitelist is not None:
            self.domain_whitelist = whitelist

    def validate(self, value):
        if not value or '@' not in value:
            raise ValidationError(self.message)

        user_part, domain_part = value.rsplit('@', 1)

        if not self.user_regex.match(user_part):
            raise ValidationError(message=self.message)

        if (domain_part not in self.domain_whitelist and
                not self.validate_domain_part(domain_part)):
            # Try for possible IDN domain-part
            try:
                domain_part = domain_part.encode('idna').decode('ascii')
            except UnicodeError:
                pass
            else:
                if self.validate_domain_part(domain_part):
                    return
            raise ValidationError(self.message)

    def validate_domain_part(self, domain_part):
        if self.domain_regex.match(domain_part):
            return True

        literal_match = self.literal_regex.match(domain_part)
        if literal_match:
            ip_address = literal_match.group(1)
            try:
                validate_ipv46_address(ip_address)
                return True
            except ValidationError:
                pass
        return False

register('email', EmailValidator)


class URLValidator(Validator):
    ul = '\u00a1-\uffff'  # unicode letters range (must not be a raw string)

    # IP patterns
    ipv4_re = (r'(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)'
               r'(?:\.(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}')
    ipv6_re = r'\[[0-9a-f:\.]+\]'  # (simple regex, validated later)

    # Host patterns
    hostname_re = (r'[a-z' + ul + r'0-9]'
                   r'(?:[a-z' + ul + r'0-9-]{0,61}[a-z' + ul + r'0-9])?')
    # Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
    domain_re = r'(?:\.(?!-)[a-z' + ul + r'0-9-]{1,63}(?<!-))*'
    tld_re = (
        r'\.'                                # dot
        r'(?!-)'                             # can't start with a dash
        r'(?:[a-z' + ul + '-]{2,63}'         # domain label
        r'|xn--[a-z0-9]{1,59})'              # or punycode label
        r'(?<!-)'                            # can't end with a dash
        r'\.?'                               # may have a trailing dot
    )
    host_re = '(' + hostname_re + domain_re + tld_re + '|localhost)'

    regex = re.compile(
        r'^(?:[a-z0-9\.\-\+]*)://'  # scheme is validated separately
        r'(?:[^\s:@/]+(?::[^\s:@/]*)?@)?'  # user:pass authentication
        r'(?:' + ipv4_re + '|' + ipv6_re + '|' + host_re + ')'
        r'(?::\d{2,5})?'  # port
        r'(?:[/?#][^\s]*)?'  # resource path
        r'\Z', re.IGNORECASE)
    schemes = ['http', 'https', 'ftp', 'ftps']
    message = 'the URL is invalid'

    def __init__(self, schemes=None, **kwargs):
        """
        Initialise the ``url`` validator.

        :param schemes: an optional list of url schemes
        :type schemes: list
        :param message: the error message in case that validation fails.
        :type message: str
        """
        super().__init__(**kwargs)
        if schemes is not None:
            self.schemes = schemes

    def validate(self, value):
        scheme = value.split('://')[0].lower()
        if scheme not in self.schemes:
            raise ValidationError(message=self.message)

        # Then check full URL
        try:
            super().validate(value)
        except ValidationError as e:
            # Trivial case failed. Try for possible IDN domain
            if value:
                try:
                    scheme, netloc, path, query, fragment = urlsplit(value)
                except ValueError:  # for example, "Invalid IPv6 URL"
                    raise ValidationError(message=self.message)
                try:
                    netloc = netloc.encode('idna').decode('ascii')
                except UnicodeError:  # invalid domain part
                    raise e
                url = urlunsplit((scheme, netloc, path, query, fragment))
                super().validate(url)
            else:
                raise
        else:
            # Now verify IPv6 in the netloc part
            host_match = re.search(r'^\[(.+)\](?::\d{2,5})?$',
                                   urlsplit(value).netloc)
            if host_match:
                potential_ip = host_match.groups()[0]
                try:
                    _validate_ipv6_address(potential_ip)
                except ValidationError:
                    raise ValidationError(message=self.message)

        # The maximum length of a full host name is 253 characters per RFC 1034
        # section 3.1. It's defined to be 255 bytes or less, but this includes
        # one byte for the length of the name and one byte for the trailing dot
        # that's used to indicate absolute names in DNS.
        if len(urlsplit(value).netloc) > 253:
            raise ValidationError(message=self.message)


register('url', URLValidator)


class MinLengthValidator(Validator):
    def __init__(self, min_length, message=None):
        """
        Initialise the ``min-length`` validator.

        :param min_length: Minumum length from which the value has to be
                           considered valid.
        :type min_length: int
        :param message: the error message in case that validation fails.
        :type message: str
        """

        self.min_length = min_length
        self.message = message or 'the length of this field must be at '\
            'least {} characters long'.format(self.min_length)

    def validate(self, value):
        if len(value) < self.min_length:
            raise ValidationError(message=self.message)


register('min-length', MinLengthValidator)

class MaxLengthValidator(Validator):
    def __init__(self, max_length, message=None):
        """
        Initialise the ``max-length`` validator.

        :param max_length: Maximum length from which the value has to be
                           considered invalid.
        :type max_length: int
        :param message: the error message in case that validation fails.
        :type message: str
        """
        self.max_length = max_length
        self.message = message or 'the length of this field must be at '\
            'most {} characters long'.format(self.max_length)

    def validate(self, value):
        if len(value) > self.max_length:
            raise ValidationError(message=self.message)


register('max-length', MaxLengthValidator)


class NumberValidator(Validator):
    def __init__(self, message=None):
        self.message = message or 'this field must be a number'

    def validate(self, value):
        try:
            float(value)
        except ValueError:
            raise ValidationError(message=self.message)

register('number', NumberValidator)

class IntegerValidator(Validator):
    def __init__(self, message=None):
        self.message = message or 'this field must be an integer'

    def validate(self, value):
        try:
            int(value)
        except ValueError:
            raise ValidationError(message=self.message)

register('integer', IntegerValidator)

class IPv4Validator(Validator):
    def __init__(self, message=None):
        self.message = message or 'this field must be an IPv4 address'

    def validate(self, value):
        try:
            ipaddress.IPv4Address(value)
        except ValueError:
            raise ValidationError(message=self.message)

register('ipv4', IPv4Validator)

class RangeValidator(Validator):
    def __init__(self, min=None, max=None, message=None):
        self.min = float(min)
        self.max = float(max)
        self.message = message or \
            'this field must be a number between {} and {}'.format(min, max)
    def validate(self, value):
        try:
            value = float(value)
            if value < self.min or value > self.max:
                raise ValidationError(message=self.message)
        except ValueError:
            raise ValidationError(message=self.message)


register('range', RangeValidator)

class MinValidator(Validator):
    def __init__(self, min=None, message=None):
        self.min = float(min)
        self.message = message or \
            'this field must be greater or equal to {}'.format(min)
    def validate(self, value):
        try:
            value = float(value)
            if value < self.min:
                raise ValidationError(message=self.message)
        except ValueError:
            raise ValidationError(message=self.message)

register('min', MinValidator)

class MaxValidator(Validator):
    def __init__(self, max=None, message=None):
        self.max = float(max)
        self.message = message or \
            'this field must be smaller or equal to {}'.format(max)
    def validate(self, value):
        try:
            value = float(value)
            if value > self.max:
                raise ValidationError(message=self.message)
        except ValueError:
            raise ValidationError(message=self.message)

register('max', MaxValidator)

class DateTimeValidator(Validator):
    def __init__(self, format_pattern='%Y-%m-%dT%H:%M:%S', message=None):
        self.format_pattern = format_pattern
        self.message = message or \
            'this field is not a valid datetime'

    def validate(self, value):
        if not value:
            return
        try:
            datetime.strptime(value, self.format_pattern)
        except ValueError:
            raise ValidationError(message=self.message)

register('datetime', DateTimeValidator)


class DateTimeRangeValidator(Validator):
    def __init__(self, format_pattern='%Y-%m-%d', message=None):
        self.format_pattern = format_pattern
        self.message = message or \
            'this field is not a valid datetime range'

    def validate(self, value):
        if not value:
            return
        d_from = None
        d_to = None
        try:
            if value['from']:
                d_from = datetime.strptime(value['from'], self.format_pattern)
            if value['to']:
                d_to = datetime.strptime(value['to'], self.format_pattern)
        except ValueError:
            raise ValidationError(message=self.message)

        if d_from and d_to and d_from > d_to:
            raise ValidationError(message=self.message)
        

register('datetimerange', DateTimeValidator)


from prompt_toolkit.application import Application
from prompt_toolkit.application.current import get_app, create_app_session, set_app
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.key_binding.bindings.focus import (focus_next,
                                                       focus_previous)
from prompt_toolkit.key_binding.defaults import load_key_bindings
from prompt_toolkit.key_binding.key_bindings import (KeyBindings,
                                                     merge_key_bindings)
from prompt_toolkit.layout import HorizontalAlign, HSplit, Layout, Window
from prompt_toolkit.shortcuts import message_dialog
from prompt_toolkit.widgets import Button, Dialog, Label

from .handlers import get_instance
from .themes import for_dialog
from .utils import validate_questions
from .validators import Validator
from .themes import set_theme

__all__ = ['dialogus']

def show_error_dialog(messages):
    texts = []
    for message in messages:
        texts.append(
            Label(message, style='class:error', dont_extend_height=True)
        )
    dialog = Dialog(
        title='Some inputs are invalid',
        body=HSplit(
            texts,
            padding=1
        ),
        buttons=[
            Button(text='Ok', handler=lambda: get_app().exit()),
        ],
        with_background=True)


    app = Application(
        layout=Layout(dialog),
        key_bindings=load_key_bindings(),
        mouse_support=True,
        style=for_dialog(),
        full_screen=True)

    app.run()

def show_dialog(questions, title, confirm, cancel):

    handlers = []
    layouts = []
    for q in questions:
        handler = get_instance(q)
        l = handler.get_layout()
        l.align = HorizontalAlign.JUSTIFY
        layouts.append(l)
        handlers.append(handler)

    def ok_handler():
        result = dict()
        for handler in handlers:
            result.update(handler.get_answer())
        get_app().exit(result=result)


    dialog = Dialog(
        title=title,
        body=HSplit(
            layouts,
            padding=1
        ),
        buttons=[
            Button(text=cancel, handler=lambda: get_app().exit()),
            Button(text=confirm, handler=ok_handler),
        ],
        with_background=True)

    # Key bindings.
    bindings = KeyBindings()

    app = Application(
        layout=Layout(dialog),
        key_bindings=merge_key_bindings([
            load_key_bindings(),
            bindings,
        ]),
        mouse_support=True,
        style=for_dialog(),
        full_screen=True)

    size = app.renderer.output.get_size()
    container = app.layout.container
    height = container.preferred_height(size.columns, size.rows).preferred
    # print(height, size.rows)
    # if height > size.rows:
    #     message_dialog(
    #         title='Too many questions',
    #         text='Cannot render a {} rows dialog in a '
    #              '{} rows screen: too many questions!'.format(height,
    #                                                           size.rows),
    #         ok_text='Got it!'
    #     )
    #     return

    while True:
        validation_errors = []
        answers = app.run()
        if answers is None:
            return
        for handler in handlers:
            if not handler.is_valid():
                for msg in handler.errors:
                    validation_errors.append(
                        '{}: {}'.format(
                            handler.get_variable_name(),
                            msg
                        )
                    )
        if not validation_errors:
            return answers

        show_error_dialog(validation_errors)

def dialogus(
    questions,
    title='Please fill the following form',
    confirm='Ok',
    cancel='Cancel',
    theme='default',
):
    """
    Show a dialog with inputs as defined in the questions parameter and returns
    a dictionary with the answers.

    :param questions: a list of questions.
    :type questions: list

    :param title: the title of the dialog.
    :type title: str

    :param confirm: the confirm button text.
    :type confirm: str

    :param cancel: the cancel button text.
    :type cancel: str

    :return: a dictionary with the answers.
    :rtype: dict

    Usage:

    .. code-block:: python

        from interrogatio import dialogus
        questions = [
            {
                'name': 'name',
                'type': 'input',
                'message': 'What is your name'
            },
            {
                'name': 'favorite_pet',
                'type': 'input',
                'message': 'What is your favorite pet'
            }
        ]
        answers = dialogus(
            questions,
            title='Tell me something about you',
            confirm='Done',
            cancel='Skip')
    """
    set_theme(theme)
    validate_questions(questions)
    return show_dialog(questions, title, confirm, cancel)
